/******************************************************************************
 * SOL-connect.c
 *
 * Dirty hack to login to a blade management module and connect to a console
 * waiting for a GDB to attach.  Very Xen specific
 *
 * Copyright (C) 2006 Tony Breeds, IBM Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation; or, when distributed
 * separately from the Linux kernel or incorporated into other
 * software packages, subject to the following license:
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this source file (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify,
 * merge, publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "exec.h"
#include "debug.h"

#define PROXY_PRG	"/usr/bin/telnet"

typedef uint8_t byte;

static pid_t telnet_pid = -1;
static int remote_in;
static int remote_out;
static int local_in;
static int local_out;

static struct expect {
	char *expect;
	char *send;
} expects[] = {
	{ .expect = "username: ",        .send   = "ozlabs\r\n", },
	{ .expect = "password: ",        .send   = "XXX PASSWORD XXX\r\n", },
	{ .expect = "system> ",          .send   = "env -T blade[4]\r\n", },
	{ .expect = "system:blade[4]> ", .send   = "console -o\r\n", },
	{ NULL, NULL}
};
static int next_expect = 0;
static bool needs_eol_hack = true;

static void usage(void)
{
	/* FIXME: Actually provide help! */
	fprintf(stderr, "id10t!\n");
	exit(1);
}

static void run_telnet(char *server, char *port)
{
	int ret;
	size_t len;
	char buf[4096];

	/* FIXME: is SOL 8-bit clean? */
	len = sprintf(buf, "%s -8 %s %s", PROXY_PRG, server, port);
	ret = run_cmd(buf, &remote_out, &remote_in);
	if (ret < 0)
		err(1, "Running Telnet failed");
	telnet_pid = ret;

	return;
}

static void do_eol_hack(int *in_len, uint8_t* in_buf, uint8_t *out_buf)
{
	int i,j;

	if (needs_eol_hack) {
		/* Yck insert a \r before every \n unless it's there */
		for(i=0,j=0; i<=*in_len; i++, j++) {
			if (in_buf[i] == '\n' && in_buf[i-1] != '\r') {
				out_buf[j++] = '\r';
			}
			out_buf[j] = in_buf[i];
		}
		*in_len = j;
	} else {
		memcpy(out_buf, in_buf, *in_len);
	}

	return;
}

static int data_from_user(void)
{
        bool more = true;
        uint8_t buf[4096];
        uint8_t buf2[4096];
        int in_len, out_len;
	int infd = local_in, outfd = remote_out;

        while (more) {
		memset(buf,  0, sizeof(buf) );
		memset(buf2, 0, sizeof(buf2));

                in_len = read(infd, buf, sizeof(buf));
		if (in_len < 0) {
			warn("Read from %d returned %d", infd, in_len);
			more = false;
		} else {
			more = (in_len == sizeof(buf));
			debug_log_data("Read", infd, buf, out_len);
			/* FIXME: Overrun */
			do_eol_hack(&in_len, buf, buf2);
			out_len = write(outfd, buf2, in_len);
			if (out_len != in_len)
				warn("Short write to fd(%d)", outfd);

			debug_log_data("Writing", outfd, buf2, out_len);
		}
        }

        return in_len;
}

static void do_expect(uint8_t *in_buf)
{
	int out_len;
	int send_len;
	uint8_t out_buf[4096];

	if (expects[next_expect].expect == NULL)
		return;

	/* Yck! */
	if (strstr((char*)in_buf, expects[next_expect].expect)) {
		send_len = strlen(expects[next_expect].send);

		memcpy(out_buf, expects[next_expect].send, send_len);
		out_len = write(remote_out, out_buf, send_len);
		if (out_len != send_len)
			warn("Short write in expect!");
		next_expect++;
		/* Assume when we get to the end of the expect script that
		 * we're happy */
		needs_eol_hack = (expects[next_expect].expect != NULL);
	}

	return;
}

static int data_from_remote(void)
{
        bool more = true;
	uint8_t buf[4096];
        int in_len, out_len;
	int infd = remote_in, outfd = local_out;

        while (more) {
                in_len = read(infd, buf, sizeof(buf));
		if (in_len < 0) {
			warn("Read from %d returned %d", infd, in_len);
			more = false;
		} else {
			more = (in_len == sizeof(buf));
			debug_log_data("Read", infd, buf, in_len);
			out_len = write(outfd, buf, in_len);
			if (out_len != in_len)
				warn("Short write to fd(%d)", outfd);
			debug("Wrote %d bytes to terminal(%d)", out_len, outfd);
			do_expect(buf);
			debug_log_data("Writing", outfd, buf, out_len);
		}
        }

        return in_len;
}

int main(int argc, char **argv)
{
	int opt, ret;
	fd_set fds;
	struct timeval *timeout = NULL;
	char *server = "blademm1", *port = "23";

	debug_init(".SOL-connect.log");

	while ( (opt = getopt(argc, argv, "s:p:")) != EOF) {
		switch(opt) {
		case 's':
			server = optarg;
			break;
		case 'p':
			port = optarg;
			break;
		default:
			errx(1, "Unknown ARG '%s'", optarg);
		}
	}

	/* Keep argc and argv in sync for old times sake */
	argv += optind;
	argc -= optind;
	if (argc > 0)
		usage();

	if (!server)
		usage();

	run_telnet(server, port);

	local_in = fileno(stdin);
	local_out = fileno(stdout);

#ifdef WITH_TIMEOUT
	timeout = calloc(1, sizeof(*timeout));
#endif
	setbuf(stdin, NULL);
	do {
		FD_ZERO(&fds);
		/* FIXME: Hack Alert */
		if (!needs_eol_hack)
			FD_SET(local_in, &fds);
		FD_SET(remote_in, &fds);

#ifdef WITH_TIMEOUT
		timeout->tv_sec=5;
		timeout->tv_usec=0;
#endif

		ret = select(getdtablesize(), &fds, NULL, NULL, timeout);
		/* FIXME: general robustness */
		if (ret == -1)
			err(1, "select()");
		else if (!ret)
			errx(1, "Timeout!!");

		if (FD_ISSET(local_in, &fds))
			ret = data_from_user();
		else if (FD_ISSET(remote_in, &fds))
			ret = data_from_remote();
	} while (ret != 0);

	/* FIXME: Clean up */
	return 0;
}
