/******************************************************************************
 * gdbproxy.c
 *
 * Spawn a command that will connect to a Xen machine waiting for GDB.
 * Once established, listen on port 1234 for a GDB process to conenct.
 * shvel data between the 2.
 *
 * Copyright (C) 2006 Tony Breeds, IBM Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation; or, when distributed
 * separately from the Linux kernel or incorporated into other
 * software packages, subject to the following license:
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this source file (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify,
 * merge, publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "exec.h"
#include "debug.h"

/* FIXME: This shouldn't be a compile time setting */
#ifndef DEFAULT_PROXY
#define DEFAULT_PROXY "/home/tony/bin/gdb_connect"
#endif

#define status(fmt, ...)	fprintf(stderr, fmt "\n", ## __VA_ARGS__)

static int gdb_socket = 0;
static int from_proxy = -1, to_proxy = -1;

static int data_from_gdb(void)
{
        bool more = true;
        uint8_t buf[4096];
        int in_len, out_len;
	int infd = gdb_socket, outfd = to_proxy;

        while (more) {
		memset(buf,  0, sizeof(buf) );

                in_len = read(infd, buf, sizeof(buf));
		if (in_len < 0) {
			warn("Read from %d returned %d", infd, in_len);
			more = false;
		} else {
			debug("Read %d bytes from gdb(%d)", in_len, infd);
			more = (in_len == sizeof(buf));
			debug_log_data("Read", infd, buf, in_len);

			out_len = write(outfd, buf, in_len);
			if (out_len != in_len)
				warn("Short write to fd(%d)", outfd);

			debug("Wrote %d bytes to hw(%d)", out_len, outfd);
			debug_log_data("Writing", outfd, buf, out_len);
		}
        }

        return in_len;
}


static int data_from_hw(void)
{
        bool more = true;
	uint8_t buf[4096];
        int in_len, out_len;
	int outfd = gdb_socket, infd = from_proxy;

        while (more) {
                in_len = read(infd, buf, sizeof(buf));
		debug("Read %d bytes from hw(%d)", in_len, infd);
		more = (in_len == sizeof(buf));
		if (in_len < 0) {
			warn("Read from %d returned %d", infd, in_len);
			more = false;
		} else {
			debug_log_data("Read", infd, buf, in_len);

			out_len = write(outfd, buf, in_len);
			if (out_len != in_len)
				warn("Short write to fd(%d)", outfd);

			debug("Wrote %d bytes to gdb(%d)", out_len, outfd);
			debug_log_data("Writing", outfd, buf, out_len);
		}
        }

        return in_len;
}

static int listen_on(int port)
{
	int ret;
	int sock;
	int on = 1;
	struct sockaddr_in sin;

	memset(&sin, 0, sizeof(sin));

	sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock < 0)
		err(1, "port %d socket() error", port);

	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = htons(port);

	ret = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if (ret)
		err(1, "port %d setsockopt() error", port);
	ret = bind(sock, (struct sockaddr *)&sin,  sizeof(sin));
	if (ret)
		err(1, "port %d bind() error", port);
	ret = listen(sock, 0);
	if (ret)
		err(1, "port %d listen() error", port);
	ret = fcntl(sock, F_SETFL, O_NDELAY);
	if (ret < 0)
		err(1, "port %d fcntl() error", port);

	return sock;
}

static int init_server(int port)
{
	int fd;
	int sock;
	socklen_t len;
	struct sockaddr_in sin;

	sock = listen_on(port);

	fd = -1;

	do {
		sleep(1);
		len = sizeof(sin);
		fd = accept(sock, (struct sockaddr*)&sin, &len);
	} while (fd == -1);

	return fd;
}

/* Wait for the console to be ready */
/* FIXME: Yck!  This really should be done by the proxy command */
static void wait_for_connect(void)
{
	int in_len;
	int len = 0;
	char *found;
	uint8_t buf[4096];
	char needle[] = " for help]";

	/* Build up the input from the hw until we see needle */
	do {
		in_len = read(from_proxy, &buf[len], sizeof(buf) - len);
		if (in_len < 0)
			err(1, "%s() read from hw failed", __func__);
		found = strstr((char*)buf, needle);
	} while (!found);

	return;
}

static void run_proxy(void)
{
	int ret;
	char *cmd;

	cmd = getenv("PROXY_COMMAND");
	ret = run_cmd((cmd?cmd:DEFAULT_PROXY), &to_proxy, &from_proxy);
	if (ret < 0)
		err(1, "Failed to run proxy \"%s\"", (cmd?cmd:DEFAULT_PROXY));

	/* FIXME: This need to be generic */
	/* If we're using the dafault proxy then wait for it to be ready
	 * (it's slow) */
	if (!cmd) {
		/* Give the slow console a change to esablish */
		status("Sleeping on proxy");
		sleep(10);
		status("Waiting for hardware");
		wait_for_connect();
	}

	return;
}


int main(void)
{
	int ret;
	fd_set fds;
	struct timeval *timeout = NULL;

	debug_init(".gdbproxy.log");

	status("Launching proxy");
	run_proxy();
	status("Launching Server [connect GDB now]");
	gdb_socket = init_server(1234);
	status("Ready");

	debug("fd(%d) == gdb_socket", gdb_socket);
	debug("fd(%d) == from_proxy", from_proxy);
	debug("fd(%d) == to_proxy", to_proxy);

#ifdef WITH_TIMEOUT
	timeout = calloc(1, sizeof(*timeout));
#endif
	setbuf(stdin, NULL);
	do {
		FD_ZERO(&fds);
		FD_SET(gdb_socket, &fds);
		FD_SET(from_proxy, &fds);

#ifdef WITH_TIMEOUT
		timeout->tv_sec=15;
		timeout->tv_usec=0;
#endif
		ret = select(getdtablesize(), &fds, NULL, NULL, timeout);
		/* FIXME: general robustness */
		if (ret == -1)
			err(1, "select()");
		else if (!ret)
			errx(1, "Timeout!!");

		if (FD_ISSET(gdb_socket, &fds))
			ret = data_from_gdb();
		else if (FD_ISSET(from_proxy, &fds))
			ret = data_from_hw();
	} while (ret != 0);

	/* FIXME: Clean up */
	return 0;
}
